# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import mock
import pytest
import requests

from paasta_tools import smartstack_tools
from paasta_tools.smartstack_tools import backend_is_up
from paasta_tools.smartstack_tools import DiscoveredHost
from paasta_tools.smartstack_tools import get_replication_for_services
from paasta_tools.smartstack_tools import ip_port_hostname_from_svname
from paasta_tools.smartstack_tools import match_backends_and_pods
from paasta_tools.utils import DEFAULT_SYNAPSE_HAPROXY_URL_FORMAT


def test_load_smartstack_info_for_service(system_paasta_config):
    with mock.patch(
        "paasta_tools.smartstack_tools.long_running_service_tools.load_service_namespace_config",
        autospec=True,
    ), mock.patch(
        "paasta_tools.smartstack_tools.get_smartstack_replication_for_attribute",
        autospec=True,
    ):
        # just a smoke test for now.
        smartstack_tools.load_smartstack_info_for_service(
            service="service",
            namespace="namespace",
            soa_dir="fake",
            blacklist=[],
            system_paasta_config=system_paasta_config,
        )


def test_get_smartstack_replication_for_attribute(system_paasta_config):
    fake_namespace = "fake_main"
    fake_service = "fake_service"
    mock_filtered_slaves = [
        {"hostname": "hostone", "attributes": {"fake_attribute": "foo"}},
        {"hostname": "hostone", "attributes": {"fake_attribute": "bar"}},
    ]

    with mock.patch(
        "paasta_tools.mesos_tools.get_all_slaves_for_blacklist_whitelist",
        return_value=mock_filtered_slaves,
        autospec=True,
    ) as mock_get_all_slaves_for_blacklist_whitelist, mock.patch(
        "paasta_tools.smartstack_tools.get_replication_for_services",
        return_value={},
        autospec=True,
    ) as mock_get_replication_for_services:
        expected = {"foo": {}, "bar": {}}
        actual = smartstack_tools.get_smartstack_replication_for_attribute(
            attribute="fake_attribute",
            service=fake_service,
            namespace=fake_namespace,
            blacklist=[],
            system_paasta_config=system_paasta_config,
        )
        mock_get_all_slaves_for_blacklist_whitelist.assert_called_once_with(
            blacklist=[], whitelist=None
        )
        assert actual == expected
        assert mock_get_replication_for_services.call_count == 2

        mock_get_replication_for_services.assert_any_call(
            synapse_host="hostone",
            synapse_port=system_paasta_config.get_synapse_port(),
            synapse_haproxy_url_format=system_paasta_config.get_synapse_haproxy_url_format(),
            services=["fake_service.fake_main"],
        )


def test_get_replication_for_service():
    testdir = os.path.dirname(os.path.realpath(__file__))
    testdata = os.path.join(testdir, "haproxy_snapshot.txt")
    with open(testdata, "r") as fd:
        mock_haproxy_data = fd.read()

    mock_response = mock.Mock()
    mock_response.text = mock_haproxy_data
    mock_get = mock.Mock(return_value=(mock_response))

    with mock.patch.object(requests.Session, "get", mock_get):
        replication_result = get_replication_for_services(
            "fake_host",
            6666,
            DEFAULT_SYNAPSE_HAPROXY_URL_FORMAT,
            ["service1", "service2", "service3", "service4"],
        )
        expected = {"service1": 18, "service2": 19, "service3": 0, "service4": 3}
        assert expected == replication_result


def test_backend_is_up():
    assert True is backend_is_up({"status": "UP"})
    assert True is backend_is_up({"status": "UP 1/2"})
    assert False is backend_is_up({"status": "DOWN"})
    assert False is backend_is_up({"status": "DOWN 1/2"})
    assert False is backend_is_up({"status": "MAINT"})


def test_ip_port_hostname_from_svname_new_format():
    assert ("10.40.10.155", 31219, "myhost") == ip_port_hostname_from_svname(
        "myhost_10.40.10.155:31219"
    )


def test_ip_port_hostname_from_svname_old_format():
    assert ("10.85.5.101", 3744, "myhost") == ip_port_hostname_from_svname(
        "10.85.5.101:3744_myhost"
    )


def test_match_backends_and_pods():
    backends = [
        {
            "pxname": "servicename.main",
            "svname": "10.50.2.4:31000_box4",
            "status": "UP",
        },
        {
            "pxname": "servicename.main",
            "svname": "10.50.2.5:31000_box5",
            "status": "UP",
        },
        {
            "pxname": "servicename.main",
            "svname": "10.50.2.6:31000_box6",
            "status": "UP",
        },
        {
            "pxname": "servicename.main",
            "svname": "10.50.2.8:31000_box8",
            "status": "UP",
        },
    ]

    good_pod1 = mock.Mock(status=mock.Mock(pod_ip="10.50.2.4"))
    good_pod2 = mock.Mock(status=mock.Mock(pod_ip="10.50.2.5"))
    bad_pod = mock.Mock(status=mock.Mock(pod_ip="10.50.2.7"))
    pods = [good_pod1, good_pod2, bad_pod]

    expected = [
        (backends[0], good_pod1),
        (backends[1], good_pod2),
        (None, bad_pod),
        (backends[2], None),
        (backends[3], None),
    ]
    actual = match_backends_and_pods(backends, pods)

    def keyfunc(t):
        return tuple(sorted((t[0] or {}).items())), t[1]

    assert sorted(actual, key=keyfunc) == sorted(expected, key=keyfunc)


@mock.patch("paasta_tools.smartstack_tools.get_multiple_backends", autospec=True)
def test_get_replication_for_all_services(mock_get_multiple_backends):
    mock_get_multiple_backends.return_value = [
        {
            "pxname": "servicename.main",
            "svname": "10.50.2.4:31000_box4",
            "status": "UP",
        },
        {
            "pxname": "servicename.main",
            "svname": "10.50.2.5:31001_box5",
            "status": "UP",
        },
        {
            "pxname": "servicename.main",
            "svname": "10.50.2.6:31001_box6",
            "status": "UP",
        },
        {
            "pxname": "servicename.main",
            "svname": "10.50.2.6:31002_box7",
            "status": "UP",
        },
        {
            "pxname": "servicename.main",
            "svname": "10.50.2.8:31000_box8",
            "status": "UP",
        },
    ]
    assert {"servicename.main": 5} == smartstack_tools.get_replication_for_all_services(
        "", 8888, ""
    )


def test_are_services_up_on_port():
    with mock.patch(
        "paasta_tools.smartstack_tools.get_multiple_backends", autospec=True
    ) as mock_get_multiple_backends, mock.patch(
        "paasta_tools.smartstack_tools.ip_port_hostname_from_svname", autospec=True
    ) as mock_ip_port_hostname_from_svname, mock.patch(
        "paasta_tools.smartstack_tools.backend_is_up", autospec=True
    ) as mock_backend_is_up:
        # none present
        assert not smartstack_tools.are_services_up_on_ip_port(
            synapse_host="1.2.3.4",
            synapse_port=3212,
            synapse_haproxy_url_format="thing",
            services=["service1.instance1", "service1.instance2"],
            host_ip="10.1.1.1",
            host_port=8888,
        )

        mock_get_multiple_backends.return_value = [
            {"svname": "thing1", "pxname": "service1.instance1"},
            {"svname": "thing2", "pxname": "service1.instance2"},
            {"svname": "thing3", "pxname": "service1.instance2"},
        ]
        mock_ip_port_hostname_from_svname.return_value = ("10.1.1.1", 8888, "a")
        mock_backend_is_up.return_value = True
        # all backends present and up
        assert smartstack_tools.are_services_up_on_ip_port(
            synapse_host="1.2.3.4",
            synapse_port=3212,
            synapse_haproxy_url_format="thing",
            services=["service1.instance1", "service1.instance2"],
            host_ip="10.1.1.1",
            host_port=8888,
        )

        mock_get_multiple_backends.return_value = [
            {"svname": "thing1", "pxname": "service1.instance1"},
            {"svname": "thing2", "pxname": "service1.instance2"},
            {"svname": "thing3", "pxname": "service1.instance2"},
        ]
        mock_ip_port_hostname_from_svname.return_value = ("10.1.1.1", 8888, "a")
        mock_backend_is_up.side_effect = [True, False, False]
        # all present bu both of service1.instance2 are DOWN
        assert not smartstack_tools.are_services_up_on_ip_port(
            synapse_host="1.2.3.4",
            synapse_port=3212,
            synapse_haproxy_url_format="thing",
            services=["service1.instance1", "service1.instance2"],
            host_ip="10.1.1.1",
            host_port=8888,
        )

        mock_get_multiple_backends.return_value = [
            {"svname": "thing1", "pxname": "service1.instance1"},
            {"svname": "thing2", "pxname": "service1.instance2"},
            {"svname": "thing3", "pxname": "service1.instance2"},
        ]
        mock_ip_port_hostname_from_svname.return_value = ("10.1.1.1", 8888, "a")
        mock_backend_is_up.side_effect = [True, True, False]
        # all present but 1 of service1.instance2 is UP
        assert smartstack_tools.are_services_up_on_ip_port(
            synapse_host="1.2.3.4",
            synapse_port=3212,
            synapse_haproxy_url_format="thing",
            services=["service1.instance1", "service1.instance2"],
            host_ip="10.1.1.1",
            host_port=8888,
        )

        mock_get_multiple_backends.return_value = [
            {"svname": "thing1", "pxname": "service1.instance1"},
            {"svname": "thing2", "pxname": "service1.instance2"},
            {"svname": "thing3", "pxname": "service1.instance2"},
        ]
        mock_ip_port_hostname_from_svname.return_value = ("10.1.1.1", 8888, "a")
        mock_backend_is_up.return_value = True
        mock_backend_is_up.side_effect = None
        # all up and present but service1.instance3 not present
        assert not smartstack_tools.are_services_up_on_ip_port(
            synapse_host="1.2.3.4",
            synapse_port=3212,
            synapse_haproxy_url_format="thing",
            services=["service1.instance1", "service1.instance2", "service1.instance3"],
            host_ip="10.1.1.1",
            host_port=8888,
        )


@pytest.fixture
def mock_replication_checker():
    smartstack_tools.BaseReplicationChecker.__abstractmethods__ = frozenset()
    system_paasta_config = mock.Mock()
    return smartstack_tools.BaseReplicationChecker(
        system_paasta_config=system_paasta_config,
        service_discovery_providers=[
            smartstack_tools.SmartstackServiceDiscovery(
                system_paasta_config=system_paasta_config
            )
        ],
    )


@pytest.fixture
def mock_kube_replication_checker():
    mock_nodes = [mock.Mock()]
    mock_system_paasta_config = mock.Mock()
    mock_system_paasta_config.get_service_discovery_providers.return_value = {
        "smartstack": {},
        "envoy": {},
    }
    return smartstack_tools.KubeSmartstackEnvoyReplicationChecker(
        nodes=mock_nodes,
        system_paasta_config=mock_system_paasta_config,
    )


def test_kube_get_allowed_locations_and_hosts(mock_kube_replication_checker):
    with mock.patch(
        "paasta_tools.kubernetes_tools.load_service_namespace_config", autospec=True
    ) as mock_load_service_namespace_config, mock.patch(
        "paasta_tools.kubernetes_tools.get_nodes_grouped_by_attribute", autospec=True
    ) as mock_get_nodes_grouped_by_attribute:
        mock_instance_config = mock.Mock(
            service="blah", instance="foo", soa_dir="/nail/thing"
        )
        mock_node_1 = mock.MagicMock(
            metadata=mock.MagicMock(
                labels={"yelp.com/hostname": "foo1", "yelp.com/pool": "default"}
            )
        )
        mock_node_2 = mock.MagicMock(
            metadata=mock.MagicMock(
                labels={"yelp.com/hostname": "foo2", "yelp.com/pool": "default"}
            )
        )
        mock_load_service_namespace_config.return_value = mock.Mock(
            get_discover=mock.Mock(return_value="region")
        )
        mock_get_nodes_grouped_by_attribute.return_value = {
            "us-west-1": [mock_node_1, mock_node_2]
        }
        ret = mock_kube_replication_checker.get_allowed_locations_and_hosts(
            mock_instance_config
        )
        assert ret == {
            "us-west-1": [
                DiscoveredHost(hostname="foo1", pool="default"),
                DiscoveredHost(hostname="foo2", pool="default"),
            ]
        }


def test_get_allowed_locations_and_hosts(mock_replication_checker):
    mock_replication_checker.get_allowed_locations_and_hosts(
        instance_config=mock.Mock()
    )


def test_get_replication_for_instance(mock_replication_checker):
    with mock.patch(
        "paasta_tools.smartstack_tools.BaseReplicationChecker.get_allowed_locations_and_hosts",
        autospec=True,
    ) as mock_get_allowed_locations_and_hosts, mock.patch(
        "paasta_tools.smartstack_tools.BaseReplicationChecker.get_hostnames_in_pool",
        autospec=True,
    ) as mock_get_hostnames_in_pool, mock.patch(
        "paasta_tools.smartstack_tools.BaseReplicationChecker._get_replication_info",
        autospec=True,
    ) as mock_get_replication_info:
        mock_get_hostnames_in_pool.return_value = [mock.Mock()]
        mock_get_allowed_locations_and_hosts.return_value = {
            "westeros-prod": [mock.Mock()],
            "middleearth-prod": [mock.Mock(), mock.Mock()],
        }
        expected = {
            "Smartstack": {
                "westeros-prod": mock_get_replication_info.return_value,
                "middleearth-prod": mock_get_replication_info.return_value,
            }
        }
        assert (
            mock_replication_checker.get_replication_for_instance(mock.Mock())
            == expected
        )
        assert mock_get_hostnames_in_pool.call_count == 2


def test_get_first_host_in_pool(mock_replication_checker):
    mock_host_0 = mock.Mock(hostname="host0", pool="some")
    mock_host_1 = mock.Mock(hostname="host123", pool="default")
    mock_host_2 = mock.Mock(hostname="host456", pool="default")
    mock_host_3 = mock.Mock(hostname="host789", pool="special")
    mock_hosts = [mock_host_0, mock_host_1, mock_host_2, mock_host_3]
    ret = mock_replication_checker.get_first_host_in_pool(mock_hosts, "default")
    assert ret == "host123"
    ret = mock_replication_checker.get_first_host_in_pool(mock_hosts, "special")
    assert ret == "host789"
    ret = mock_replication_checker.get_first_host_in_pool(mock_hosts, "what")
    assert ret == "host0"


def test_hostnames_in_pool(mock_replication_checker):
    mock_host_0 = mock.Mock(hostname="host0", pool="some")
    mock_host_1 = mock.Mock(hostname="host123", pool="default")
    mock_host_2 = mock.Mock(hostname="host456", pool="default")
    mock_host_3 = mock.Mock(hostname="host789", pool="special")
    mock_hosts = [mock_host_0, mock_host_1, mock_host_2, mock_host_3]
    ret = mock_replication_checker.get_hostnames_in_pool(mock_hosts, "default")
    assert ret == ["host123", "host456"]
    ret = mock_replication_checker.get_hostnames_in_pool(mock_hosts, "special")
    assert ret == ["host789"]
    ret = mock_replication_checker.get_hostnames_in_pool(mock_hosts, "what")
    assert ret == ["host0"]


def test_get_replication_info(mock_replication_checker):
    mock_replication_checker._cache = {}
    mock_instance_config = mock.Mock(service="thing", instance="main")
    mock_service_discovery_provider = mock.MagicMock()
    ret = mock_replication_checker._get_replication_info(
        "westeros-prod", "host1", mock_instance_config, mock_service_discovery_provider
    )
    assert ret == {
        "thing.main": mock_service_discovery_provider.get_replication_for_all_services.return_value[
            "thing.main"
        ]
    }
    assert mock_replication_checker._cache == {
        (
            "westeros-prod",
            mock_service_discovery_provider.NAME,
        ): mock_service_discovery_provider.get_replication_for_all_services.return_value
    }



import mock
import pytest

from paasta_tools import firewall
from paasta_tools import iptables
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import NoConfigurationForServiceError


EMPTY_RULE = iptables.Rule(
    protocol="ip",
    src="0.0.0.0/0.0.0.0",
    dst="0.0.0.0/0.0.0.0",
    target=None,
    matches=(),
    target_parameters=(),
)


@pytest.fixture
def service_group():
    return firewall.ServiceGroup(service="my_cool_service", instance="web")


@pytest.fixture
def mock_get_running_mesos_docker_containers():
    with mock.patch.object(
        firewall,
        "get_running_mesos_docker_containers",
        autospec=True,
        return_value=[
            {
                "HostConfig": {"NetworkMode": "bridge"},
                "Labels": {
                    "paasta_service": "myservice",
                    "paasta_instance": "hassecurity",
                },
                "NetworkSettings": {
                    "Networks": {
                        "bridge": {
                            "MacAddress": "02:42:a9:fe:00:0a",
                            "IPAddress": "1.1.1.1",
                        }
                    }
                },
            },
            {
                "HostConfig": {"NetworkMode": "bridge"},
                "Labels": {
                    "paasta_service": "myservice",
                    "paasta_instance": "hassecurity2",
                },
                "NetworkSettings": {
                    "Networks": {
                        "bridge": {
                            "MacAddress": "02:42:a9:fe:00:0b",
                            "IPAddress": "2.2.2.2",
                        }
                    }
                },
            },
            # host networking
            {
                "HostConfig": {"NetworkMode": "host"},
                "Labels": {"paasta_service": "myservice", "paasta_instance": "batch"},
            },
            # no labels
            {"HostConfig": {"NetworkMode": "bridge"}, "Labels": {}},
        ],
    ):
        yield


@pytest.mark.usefixtures("mock_get_running_mesos_docker_containers")
def test_services_running_here():
    assert tuple(firewall.services_running_here()) == (
        ("myservice", "hassecurity", "02:42:a9:fe:00:0a", "1.1.1.1"),
        ("myservice", "hassecurity2", "02:42:a9:fe:00:0b", "2.2.2.2"),
    )


@pytest.fixture
def mock_services_running_here():
    with mock.patch.object(
        firewall,
        "services_running_here",
        autospec=True,
        side_effect=lambda: iter(
            (
                ("example_happyhour", "main", "02:42:a9:fe:00:00", "1.1.1.1"),
                ("example_happyhour", "main", "02:42:a9:fe:00:01", "2.2.2.2"),
                ("example_happyhour", "batch", "02:42:a9:fe:00:02", "3.3.3.3"),
                ("my_cool_service", "web", "02:42:a9:fe:00:03", "4.4.4.4"),
                ("my_cool_service", "web", "02:42:a9:fe:00:04", "5.5.5.5"),
            )
        ),
    ):
        yield


def test_service_group_chain_name(service_group):
    """The chain name must be stable, unique, and short."""
    assert service_group.chain_name == "PAASTA.my_cool_se.f031797563"
    assert len(service_group.chain_name) <= 28


@pytest.fixture
def mock_service_config():
    with mock.patch.object(
        firewall, "get_instance_config", autospec=True
    ) as mock_instance_config, mock.patch.object(
        firewall,
        "get_all_namespaces_for_service",
        autospec=True,
        return_value={"example_happyhour.main": {"proxy_port": "20000"}},
    ), mock.patch.object(
        firewall, "load_system_paasta_config", autospec=True
    ) as mock_system_paasta_config, mock.patch.object(
        firewall, "_synapse_backends", autospec=True
    ) as mock_synapse_backends:

        mock_system_paasta_config.return_value.get_cluster.return_value = "mycluster"

        mock_instance_config.return_value = mock.Mock()
        mock_instance_config.return_value.get_dependencies.return_value = [
            {"well-known": "internet"},
            {"smartstack": "example_happyhour.main"},
            {"cidr": "169.229.226.0/24"},
            {"cidr": "8.8.8.8", "port": 53},
            {"cidr": "8.8.4.4", "port": "1024:65535"},
        ]
        mock_instance_config.return_value.get_outbound_firewall.return_value = "monitor"

        mock_synapse_backends.return_value = [
            {"host": "1.2.3.4", "port": 123},
            {"host": "5.6.7.8", "port": 567},
        ]
        yield mock_instance_config


def test_service_group_rules_no_dependencies(mock_service_config, service_group):
    mock_service_config.return_value.get_dependencies.return_value = None
    assert service_group.get_rules(
        DEFAULT_SOA_DIR, firewall.DEFAULT_SYNAPSE_SERVICE_DIR
    ) == (
        EMPTY_RULE._replace(
            target="LOG",
            matches=(("limit", (("limit", ("1/sec",)), ("limit-burst", ("1",)))),),
            target_parameters=(("log-prefix", ("paasta.my_cool_service ",)),),
        ),
        EMPTY_RULE._replace(target="PAASTA-COMMON"),
    )


def test_service_group_rules_monitor(mock_service_config, service_group):
    assert service_group.get_rules(
        DEFAULT_SOA_DIR, firewall.DEFAULT_SYNAPSE_SERVICE_DIR
    ) == (
        EMPTY_RULE._replace(
            target="LOG",
            matches=(("limit", (("limit", ("1/sec",)), ("limit-burst", ("1",)))),),
            target_parameters=(("log-prefix", ("paasta.my_cool_service ",)),),
        ),
        EMPTY_RULE._replace(target="PAASTA-COMMON"),
        EMPTY_RULE._replace(target="PAASTA-INTERNET"),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="1.2.3.4/255.255.255.255",
            matches=(
                ("comment", (("comment", ("backend example_happyhour.main",)),)),
                ("tcp", (("dport", ("123",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="5.6.7.8/255.255.255.255",
            matches=(
                ("comment", (("comment", ("backend example_happyhour.main",)),)),
                ("tcp", (("dport", ("567",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="169.254.255.254/255.255.255.255",
            matches=(
                ("comment", (("comment", ("proxy_port example_happyhour.main",)),)),
                ("tcp", (("dport", ("20000",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="ip",
            target="ACCEPT",
            dst="169.229.226.0/255.255.255.0",
            matches=(("comment", (("comment", ("allow 169.229.226.0/24:*",)),)),),
        ),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="8.8.8.8/255.255.255.255",
            matches=(
                ("comment", (("comment", ("allow 8.8.8.8/32:53",)),)),
                ("tcp", (("dport", ("53",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="udp",
            target="ACCEPT",
            dst="8.8.8.8/255.255.255.255",
            matches=(
                ("comment", (("comment", ("allow 8.8.8.8/32:53",)),)),
                ("udp", (("dport", ("53",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="8.8.4.4/255.255.255.255",
            matches=(
                ("comment", (("comment", ("allow 8.8.4.4/32:1024:65535",)),)),
                ("tcp", (("dport", ("1024:65535",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="udp",
            target="ACCEPT",
            dst="8.8.4.4/255.255.255.255",
            matches=(
                ("comment", (("comment", ("allow 8.8.4.4/32:1024:65535",)),)),
                ("udp", (("dport", ("1024:65535",)),)),
            ),
        ),
    )


def test_service_group_rules_block(mock_service_config, service_group):
    mock_service_config.return_value.get_outbound_firewall.return_value = "block"
    assert service_group.get_rules(
        DEFAULT_SOA_DIR, firewall.DEFAULT_SYNAPSE_SERVICE_DIR
    ) == (
        EMPTY_RULE._replace(
            target="REJECT",
            matches=(),
            target_parameters=(("reject-with", ("icmp-port-unreachable",)),),
        ),
        EMPTY_RULE._replace(
            target="LOG",
            matches=(("limit", (("limit", ("1/sec",)), ("limit-burst", ("1",)))),),
            target_parameters=(("log-prefix", ("paasta.my_cool_service ",)),),
        ),
        EMPTY_RULE._replace(target="PAASTA-COMMON"),
        EMPTY_RULE._replace(target="PAASTA-INTERNET"),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="1.2.3.4/255.255.255.255",
            matches=(
                ("comment", (("comment", ("backend example_happyhour.main",)),)),
                ("tcp", (("dport", ("123",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="5.6.7.8/255.255.255.255",
            matches=(
                ("comment", (("comment", ("backend example_happyhour.main",)),)),
                ("tcp", (("dport", ("567",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="169.254.255.254/255.255.255.255",
            matches=(
                ("comment", (("comment", ("proxy_port example_happyhour.main",)),)),
                ("tcp", (("dport", ("20000",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="ip",
            target="ACCEPT",
            dst="169.229.226.0/255.255.255.0",
            matches=(("comment", (("comment", ("allow 169.229.226.0/24:*",)),)),),
        ),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="8.8.8.8/255.255.255.255",
            matches=(
                ("comment", (("comment", ("allow 8.8.8.8/32:53",)),)),
                ("tcp", (("dport", ("53",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="udp",
            target="ACCEPT",
            dst="8.8.8.8/255.255.255.255",
            matches=(
                ("comment", (("comment", ("allow 8.8.8.8/32:53",)),)),
                ("udp", (("dport", ("53",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="8.8.4.4/255.255.255.255",
            matches=(
                ("comment", (("comment", ("allow 8.8.4.4/32:1024:65535",)),)),
                ("tcp", (("dport", ("1024:65535",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            protocol="udp",
            target="ACCEPT",
            dst="8.8.4.4/255.255.255.255",
            matches=(
                ("comment", (("comment", ("allow 8.8.4.4/32:1024:65535",)),)),
                ("udp", (("dport", ("1024:65535",)),)),
            ),
        ),
    )


def test_service_group_rules_synapse_backend_error(mock_service_config, service_group):
    mock_service_config.return_value.get_dependencies.return_value = [
        {"smartstack": "example_happyhour.main"}
    ]
    firewall._synapse_backends.side_effect = IOError("Error loading file")
    assert service_group.get_rules(
        DEFAULT_SOA_DIR, firewall.DEFAULT_SYNAPSE_SERVICE_DIR
    ) == (
        EMPTY_RULE._replace(
            target="LOG",
            matches=(("limit", (("limit", ("1/sec",)), ("limit-burst", ("1",)))),),
            target_parameters=(("log-prefix", ("paasta.my_cool_service ",)),),
        ),
        EMPTY_RULE._replace(target="PAASTA-COMMON"),
        EMPTY_RULE._replace(
            protocol="tcp",
            target="ACCEPT",
            dst="169.254.255.254/255.255.255.255",
            matches=(
                ("comment", (("comment", ("proxy_port example_happyhour.main",)),)),
                ("tcp", (("dport", ("20000",)),)),
            ),
        ),
    )


def test_service_group_rules_empty_when_invalid_instance_type(
    service_group, mock_service_config
):
    with mock.patch.object(
        firewall, "get_instance_config", side_effect=NotImplementedError()
    ):
        assert (
            service_group.get_rules(
                DEFAULT_SOA_DIR, firewall.DEFAULT_SYNAPSE_SERVICE_DIR
            )
            == ()
        )


def test_service_group_rules_empty_when_service_is_deleted(
    service_group, mock_service_config
):
    """A deleted service which still has running containers shouldn't cause exceptions."""
    with mock.patch.object(
        firewall, "get_instance_config", side_effect=NoConfigurationForServiceError()
    ):
        assert (
            service_group.get_rules(
                DEFAULT_SOA_DIR, firewall.DEFAULT_SYNAPSE_SERVICE_DIR
            )
            == ()
        )


@mock.patch.object(iptables, "ensure_chain", autospec=True)
@mock.patch.object(iptables, "reorder_chain", autospec=True)
def test_service_group_update_rules(reorder_mock, ensure_mock, service_group):
    with mock.patch.object(
        type(service_group), "get_rules", return_value=mock.sentinel.RULES
    ):
        service_group.update_rules(
            DEFAULT_SOA_DIR, firewall.DEFAULT_SYNAPSE_SERVICE_DIR
        )
    ensure_mock.assert_called_once_with(service_group.chain_name, mock.sentinel.RULES)
    reorder_mock.assert_called_once_with(service_group.chain_name)


def test_active_service_groups(mock_service_config, mock_services_running_here):
    assert firewall.active_service_groups() == {
        firewall.ServiceGroup("example_happyhour", "main"): {
            "02:42:a9:fe:00:00",
            "02:42:a9:fe:00:01",
        },
        firewall.ServiceGroup("example_happyhour", "batch"): {"02:42:a9:fe:00:02"},
        firewall.ServiceGroup("my_cool_service", "web"): {
            "02:42:a9:fe:00:03",
            "02:42:a9:fe:00:04",
        },
    }


def test_ensure_internet_chain():
    with mock.patch.object(iptables, "ensure_chain", autospec=True) as m:
        firewall._ensure_internet_chain()
    (call,) = m.call_args_list
    args, _ = call
    assert args[0] == "PAASTA-INTERNET"
    assert args[1] == (
        EMPTY_RULE._replace(target="ACCEPT"),
        EMPTY_RULE._replace(dst="127.0.0.0/255.0.0.0", target="RETURN"),
        EMPTY_RULE._replace(dst="10.0.0.0/255.0.0.0", target="RETURN"),
        EMPTY_RULE._replace(dst="172.16.0.0/255.240.0.0", target="RETURN"),
        EMPTY_RULE._replace(dst="192.168.0.0/255.255.0.0", target="RETURN"),
        EMPTY_RULE._replace(dst="169.254.0.0/255.255.0.0", target="RETURN"),
    )


@pytest.fixture
def mock_active_service_groups():
    groups = {
        firewall.ServiceGroup("cool_service", "main"): {
            "02:42:a9:fe:00:02",
            "fe:a3:a3:da:2d:51",
            "fe:a3:a3:da:2d:50",
        },
        firewall.ServiceGroup("cool_service", "main"): {"fe:a3:a3:da:2d:40"},
        firewall.ServiceGroup("dumb_service", "other"): {
            "fe:a3:a3:da:2d:30",
            "fe:a3:a3:da:2d:31",
        },
    }
    return groups


@mock.patch.object(iptables, "reorder_chain", autospec=True)
def test_ensure_service_chains(
    mock_reorder_chain, mock_active_service_groups, mock_service_config
):
    with mock.patch.object(iptables, "ensure_chain", autospec=True) as m:
        assert firewall.ensure_service_chains(
            mock_active_service_groups,
            DEFAULT_SOA_DIR,
            firewall.DEFAULT_SYNAPSE_SERVICE_DIR,
        ) == {
            "PAASTA.cool_servi.397dba3c1f": {"fe:a3:a3:da:2d:40"},
            "PAASTA.dumb_servi.8fb64b4f63": {"fe:a3:a3:da:2d:30", "fe:a3:a3:da:2d:31"},
        }
    assert len(m.mock_calls) == 2
    assert mock.call("PAASTA.cool_servi.397dba3c1f", mock.ANY) in m.mock_calls
    assert mock.call("PAASTA.dumb_servi.8fb64b4f63", mock.ANY) in m.mock_calls
    assert mock.call("PAASTA.cool_servi.397dba3c1f") in mock_reorder_chain.mock_calls
    assert mock.call("PAASTA.dumb_servi.8fb64b4f63") in mock_reorder_chain.mock_calls


def test_ensure_dispatch_chains():
    with mock.patch.object(
        iptables, "ensure_rule", autospec=True
    ) as mock_ensure_rule, mock.patch.object(
        iptables, "ensure_chain", autospec=True
    ) as mock_ensure_chain:
        firewall.ensure_dispatch_chains(
            {"chain1": {"mac1", "mac2"}, "chain2": {"mac3"}}
        )

    assert mock_ensure_chain.mock_calls == [
        mock.call(
            "PAASTA",
            {
                EMPTY_RULE._replace(
                    target="chain1", matches=(("mac", (("mac-source", ("MAC1",)),)),)
                ),
                EMPTY_RULE._replace(
                    target="chain1", matches=(("mac", (("mac-source", ("MAC2",)),)),)
                ),
                EMPTY_RULE._replace(
                    target="chain2", matches=(("mac", (("mac-source", ("MAC3",)),)),)
                ),
            },
        )
    ]

    assert mock_ensure_rule.mock_calls == [
        mock.call("INPUT", EMPTY_RULE._replace(target="PAASTA")),
        mock.call("FORWARD", EMPTY_RULE._replace(target="PAASTA")),
    ]


def test_garbage_collect_old_service_chains():
    with mock.patch.object(
        iptables, "delete_chain", autospec=True
    ) as mock_delete_chain, mock.patch.object(
        iptables,
        "all_chains",
        autospec=True,
        return_value={
            "INPUT",
            "OUTPUT",
            "FORWARD",
            "DOCKER",
            "PAASTA",
            "PAASTA-INTERNET",
            "PAASTA.chain1",
            "PAASTA.chain3",
        },
    ):
        firewall.garbage_collect_old_service_chains(
            {"PAASTA.chain1": {"mac1", "mac2"}, "PAASTA.chain2": {"mac3"}}
        )

    assert mock_delete_chain.mock_calls == [mock.call("PAASTA.chain3")]


@mock.patch.object(firewall.ServiceGroup, "get_rules", return_value=mock.sentinel.RULES)
@mock.patch.object(iptables, "reorder_chain", autospec=True)
@mock.patch.object(iptables, "ensure_chain", autospec=True)
@mock.patch.object(iptables, "insert_rule", autospec=True)
def test_prepare_new_container(
    insert_rule_mock, ensure_chain_mock, reorder_chain_mock, get_rules_mock
):
    firewall.prepare_new_container(
        DEFAULT_SOA_DIR,
        firewall.DEFAULT_SYNAPSE_SERVICE_DIR,
        "myservice",
        "myinstance",
        "00:00:00:00:00:00",
    )
    assert ensure_chain_mock.mock_calls == [
        mock.call("PAASTA-DNS", mock.ANY),
        mock.call("PAASTA-INTERNET", mock.ANY),
        mock.call("PAASTA-COMMON", mock.ANY),
        mock.call("PAASTA.myservice.7e8522249a", mock.sentinel.RULES),
    ]
    assert reorder_chain_mock.mock_calls == [mock.call("PAASTA.myservice.7e8522249a")]
    assert insert_rule_mock.mock_calls == [
        mock.call(
            "PAASTA",
            EMPTY_RULE._replace(
                target="PAASTA.myservice.7e8522249a",
                matches=(("mac", (("mac-source", ("00:00:00:00:00:00",)),)),),
            ),
        )
    ]


@pytest.mark.parametrize(
    ("resolv_conf", "expected"),
    (
        (
            "nameserver         8.8.8.8\n"
            "nameserver\t8.8.4.4\n"
            "nameserver 169.254.255.254\n",
            ("8.8.8.8", "8.8.4.4", "169.254.255.254"),
        ),
        (
            "#nameserver 8.8.8.8\n"
            "nameserver\n"
            "nameserver 8.8.4.4\n"
            "nameserver 2001:4860:4860::8888\n",
            ("8.8.4.4",),
        ),
        (
            "domain yelpcorp.com\n" "nameserver 8.8.4.4\n" "search a b c d\n",
            ("8.8.4.4",),
        ),
    ),
)
def test_dns_servers(tmpdir, resolv_conf, expected):
    path = tmpdir.join("resolv.conf")
    path.write(resolv_conf)
    with mock.patch.object(firewall, "RESOLV_CONF", path.strpath):
        assert tuple(firewall._dns_servers()) == expected


def test_ensure_dns_chain(tmpdir):
    path = tmpdir.join("resolv.conf")
    path.write("nameserver 8.8.8.8\n" "nameserver 8.8.4.4\n")
    with mock.patch.object(
        iptables, "ensure_chain", autospec=True
    ) as m, mock.patch.object(firewall, "RESOLV_CONF", path.strpath):
        firewall._ensure_dns_chain()
    (call,) = m.call_args_list
    args, _ = call
    assert args[0] == "PAASTA-DNS"
    assert args[1] == (
        EMPTY_RULE._replace(
            dst="8.8.8.8/255.255.255.255",
            target="ACCEPT",
            protocol="udp",
            matches=(("udp", (("dport", ("53",)),)),),
        ),
        EMPTY_RULE._replace(
            dst="8.8.8.8/255.255.255.255",
            target="ACCEPT",
            protocol="tcp",
            matches=(("tcp", (("dport", ("53",)),)),),
        ),
        EMPTY_RULE._replace(
            dst="8.8.4.4/255.255.255.255",
            target="ACCEPT",
            protocol="udp",
            matches=(("udp", (("dport", ("53",)),)),),
        ),
        EMPTY_RULE._replace(
            dst="8.8.4.4/255.255.255.255",
            target="ACCEPT",
            protocol="tcp",
            matches=(("tcp", (("dport", ("53",)),)),),
        ),
    )


def test_ensure_common_chain():
    with mock.patch.object(iptables, "ensure_chain", autospec=True) as m:
        firewall._ensure_common_chain()
    (call,) = m.call_args_list
    args, _ = call
    assert args[0] == "PAASTA-COMMON"
    assert args[1] == (
        EMPTY_RULE._replace(
            target="ACCEPT", matches=(("conntrack", (("ctstate", ("ESTABLISHED",)),)),)
        ),
        EMPTY_RULE._replace(
            dst="169.254.255.254/255.255.255.255",
            target="ACCEPT",
            protocol="tcp",
            matches=(
                ("comment", (("comment", ("scribed",)),)),
                ("tcp", (("dport", ("1463",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            dst="169.254.255.254/255.255.255.255",
            target="ACCEPT",
            protocol="udp",
            matches=(
                ("comment", (("comment", ("metrics-relay",)),)),
                ("udp", (("dport", ("8125",)),)),
            ),
        ),
        EMPTY_RULE._replace(
            dst="169.254.255.254/255.255.255.255",
            target="ACCEPT",
            protocol="tcp",
            matches=(
                ("comment", (("comment", ("sensu",)),)),
                ("tcp", (("dport", ("3030",)),)),
            ),
        ),
        EMPTY_RULE._replace(target="PAASTA-DNS"),
    )



import contextlib
import fcntl
import itertools
import os
import signal
import sys
import time

import mock
import pytest

from paasta_tools import mac_address


skip_if_osx = pytest.mark.skipif(
    sys.platform == "darwin", reason="Flock is not present on OS X"
)


def test_simple(tmpdir):
    mac, lock_file = mac_address.reserve_unique_mac_address(str(tmpdir))
    with contextlib.closing(lock_file):
        assert lock_file is not None
        assert mac == "02:52:00:00:00:00"
        assert tmpdir.join(mac).check()


def test_dir_not_exist(tmpdir):
    with pytest.raises(IOError):
        mac_address.reserve_unique_mac_address(str(tmpdir.join("nonexistent")))


def test_file_exists_no_flock(tmpdir):
    tmpdir.join("02:52:00:00:00:00").ensure()
    mac, lock_file = mac_address.reserve_unique_mac_address(str(tmpdir))
    with contextlib.closing(lock_file):
        assert lock_file is not None
        assert mac == "02:52:00:00:00:00"


def _flock_process(path):
    # fork a subprocess that holds an flock.
    r, w = os.pipe()
    child_pid = os.fork()
    if child_pid == 0:  # pragma: no cover
        os.close(r)
        fd = os.open(path, os.O_CREAT | os.O_RDWR)
        fcntl.flock(fd, fcntl.LOCK_EX)
        os.write(w, b"ok")
        os.close(w)
        time.sleep(60 * 60 * 24)  # sleep for some approximation of infinity
        sys.exit(0)  # never returns

    os.close(w)
    # wait for something to be printed so we know the flock has occurred
    assert os.read(r, 2) == b"ok"
    os.close(r)
    return child_pid


@skip_if_osx
def test_file_exists_flock(tmpdir):
    # it doesn't count if this process has the flock, so we need to spawn a different one to hold it
    flock_process = _flock_process(str(tmpdir.join("02:52:00:00:00:00")))
    try:
        mac, lock_file = mac_address.reserve_unique_mac_address(str(tmpdir))
        with contextlib.closing(lock_file):
            assert lock_file is not None
            assert mac == "02:52:00:00:00:01"
    finally:
        os.kill(flock_process, signal.SIGKILL)


@pytest.fixture(autouse=True)
def mock_randbits():
    # make getrandbits() reliably return an incrementing counter starting at 0
    class counter(itertools.count):
        def __call__(self, _):
            return next(self)

    with mock.patch.object(mac_address.random, "getrandbits", side_effect=counter()):
        yield



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mock
import pytest

from paasta_tools import check_kubernetes_services_replication
from paasta_tools import check_services_replication_tools
from paasta_tools.utils import compose_job_id

check_kubernetes_services_replication.log = mock.Mock()
check_services_replication_tools.log = mock.Mock()


@pytest.fixture
def instance_config():
    service = "fake_service"
    instance = "fake_instance"
    job_id = compose_job_id(service, instance)
    mock_instance_config = mock.Mock(
        service=service,
        instance=instance,
        cluster="fake_cluster",
        soa_dir="fake_soa_dir",
        job_id=job_id,
        config_dict={},
    )
    mock_instance_config.get_replication_crit_percentage.return_value = 90
    mock_instance_config.get_registrations.return_value = [job_id]
    return mock_instance_config


def test_check_service_replication_for_normal_smartstack(instance_config):
    instance_config.get_instances.return_value = 100
    with mock.patch(
        "paasta_tools.check_kubernetes_services_replication.get_proxy_port_for_instance",
        autospec=True,
        return_value=666,
    ), mock.patch(
        "paasta_tools.monitoring_tools.check_replication_for_instance",
        autospec=True,
    ) as mock_check_replication_for_service:
        check_kubernetes_services_replication.check_kubernetes_pod_replication(
            instance_config=instance_config,
            pods_by_service_instance={},
            replication_checker=None,
            dry_run=True,
        )
        mock_check_replication_for_service.assert_called_once_with(
            instance_config=instance_config,
            expected_count=100,
            replication_checker=None,
            dry_run=True,
        )


def test_check_service_replication_for_smartstack_with_different_namespace(
    instance_config,
):
    instance_config.get_instances.return_value = 100
    with mock.patch(
        "paasta_tools.check_kubernetes_services_replication.get_proxy_port_for_instance",
        autospec=True,
        return_value=666,
    ), mock.patch(
        "paasta_tools.monitoring_tools.check_replication_for_instance",
        autospec=True,
    ) as mock_check_replication_for_service, mock.patch(
        "paasta_tools.check_kubernetes_services_replication.check_healthy_kubernetes_tasks_for_service_instance",
        autospec=True,
    ) as mock_check_healthy_kubernetes_tasks:
        instance_config.get_registrations.return_value = ["some-random-other-namespace"]
        check_kubernetes_services_replication.check_kubernetes_pod_replication(
            instance_config=instance_config,
            pods_by_service_instance={},
            replication_checker=None,
            dry_run=True,
        )
        assert not mock_check_replication_for_service.called
        mock_check_healthy_kubernetes_tasks.assert_called_once_with(
            instance_config=instance_config,
            expected_count=100,
            pods_by_service_instance={},
            dry_run=True,
        )


def test_check_service_replication_for_non_smartstack(instance_config):
    instance_config.get_instances.return_value = 100

    with mock.patch(
        "paasta_tools.check_kubernetes_services_replication.get_proxy_port_for_instance",
        autospec=True,
        return_value=None,
    ), mock.patch(
        "paasta_tools.check_kubernetes_services_replication.check_healthy_kubernetes_tasks_for_service_instance",
        autospec=True,
    ) as mock_check_healthy_kubernetes_tasks:
        check_kubernetes_services_replication.check_kubernetes_pod_replication(
            instance_config=instance_config,
            pods_by_service_instance={},
            replication_checker=None,
            dry_run=True,
        )
        mock_check_healthy_kubernetes_tasks.assert_called_once_with(
            instance_config=instance_config,
            expected_count=100,
            pods_by_service_instance={},
            dry_run=True,
        )


def test_check_healthy_kubernetes_tasks_for_service_instance():
    with mock.patch(
        "paasta_tools.check_kubernetes_services_replication.is_pod_ready",
        autospec=True,
        side_effect=[True, False],
    ), mock.patch(
        "paasta_tools.monitoring_tools.send_replication_event_if_under_replication",
        autospec=True,
    ) as mock_send_replication_event_if_under_replication:
        mock_instance_config = mock.Mock()
        mock_pod_1 = mock.Mock()
        mock_pod_2 = mock.Mock()
        pods_by_service_instance = {
            mock_instance_config.service: {
                mock_instance_config.instance: [mock_pod_1, mock_pod_2]
            }
        }
        check_kubernetes_services_replication.check_healthy_kubernetes_tasks_for_service_instance(
            mock_instance_config,
            5,
            pods_by_service_instance=pods_by_service_instance,
            dry_run=True,
        )
        mock_send_replication_event_if_under_replication.assert_called_with(
            instance_config=mock_instance_config,
            expected_count=5,
            num_available=1,
            dry_run=True,
        )



# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import pytest
from mock import MagicMock
from mock import Mock
from mock import patch

from paasta_tools.oom_logger import capture_oom_events_from_stdin
from paasta_tools.oom_logger import log_to_clog
from paasta_tools.oom_logger import LogLine
from paasta_tools.oom_logger import main
from paasta_tools.oom_logger import send_sfx_event


@pytest.fixture
def sys_stdin():
    return [
        "some random line1\n",
        "1500316299 dev37-devc [30533610.306528] apache2 invoked oom-killer: "
        "gfp_mask=0x24000c0, order=0, oom_score_adj=0\n,"
        "some random line2\n",
        "1500316300 dev37-devc [30533610.306529] Task in "
        "/docker/a687af92e281725daf5b4cda0b487f20d2055d2bb6814b76d0e39c18a52a4e79 "
        "killed as a result of limit of "
        "/docker/a687af92e281725daf5b4cda0b487f20d2055d2bb6814b76d0e39c18a52a4e79\n",
    ]


@pytest.fixture
def sys_stdin_kubernetes_burstable_qos():
    return [
        "some random line1\n",
        "1500316299 dev37-devc [30533610.306528] apache2 invoked oom-killer: "
        "gfp_mask=0x24000c0, order=0, oom_score_adj=0\n",
        "some random line2\n",
        "1500316300 dev37-devc [30533610.306529] Task in "
        "/kubepods/burstable/podf91e9681-4741-4ef4-8f5a-182c5683df8b/"
        "0e4a814eda03622476ff47871e6c397e5b8747af209b44f3b3e1c5289b0f9772 "
        "killed as a result of limit of /kubepods/burstable/"
        "podf91e9681-4741-4ef4-8f5a-182c5683df8b/"
        "0e4a814eda03622476ff47871e6c397e5b8747af209b44f3b3e1c5289b0f9772\n",
    ]


@pytest.fixture
def sys_stdin_kubernetes_guaranteed_qos():
    return [
        "some random line1\n",
        "1500316299 dev37-devc [30533610.306528] apache2 invoked oom-killer: "
        "gfp_mask=0x24000c0, order=0, oom_score_adj=0\n",
        "some random line2\n",
        "1500316300 dev37-devc [30533610.306529] Task in "
        "/kubepods/podf91e9681-4741-4ef4-8f5a-182c5683df8b/"
        "0e4a814eda03622476ff47871e6c397e5b8747af209b44f3b3e1c5289b0f9772 "
        "killed as a result of limit of /kubepods/"
        "podf91e9681-4741-4ef4-8f5a-182c5683df8b/"
        "0e4a814eda03622476ff47871e6c397e5b8747af209b44f3b3e1c5289b0f9772\n",
    ]


@pytest.fixture
def sys_stdin_kubernetes_besteffort_qos():
    return [
        "some random line1\n",
        "1500316299 dev37-devc [30533610.306528] apache2 invoked oom-killer: "
        "gfp_mask=0x24000c0, order=0, oom_score_adj=0\n",
        "some random line2\n",
        "1500316300 dev37-devc [30533610.306529] Task in "
        "/kubepods/besteffort/podf91e9681-4741-4ef4-8f5a-182c5683df8b/"
        "0e4a814eda03622476ff47871e6c397e5b8747af209b44f3b3e1c5289b0f9772 "
        "killed as a result of limit of /kubepods/besteffort/"
        "podf91e9681-4741-4ef4-8f5a-182c5683df8b/"
        "0e4a814eda03622476ff47871e6c397e5b8747af209b44f3b3e1c5289b0f9772\n",
    ]


@pytest.fixture
def sys_stdin_kubernetes_structured_burstable_qos():
    return [
        "some random line1\n",
        "1500316299 dev37-devc [30533610.306528] apache2 invoked oom-killer: "
        "gfp_mask=0x24000c0, order=0, oom_score_adj=0\n",
        "some random line2\n",
        "1500316300 dev37-devc [541471.893603] oom-kill:constraint=CONSTRAINT_MEMCG,"
        "nodemask=(null),cpuset=1b6f55bfbed10cdc2bb6944078eaefd3278f8f9b3a9725c4ddffb722752a2279,"
        "mems_allowed=0-1,oom_memcg=/kubepods/burstable/podb56c3a7a-a84d-4f84-b97e-446c4e705259/"
        "0e4a814eda030cdc2bb6944078eaefd3278f8f9b3a9725c4ddffb722752a2279,"
        "task_memcg=/kubepods/burstable/podb56c3a7a-a84d-4f84-b97e-446c4e705259/"
        "0e4a814eda030cdc2bb6944078eaefd3278f8f9b3a9725c4ddffb722752a2279,"
        "task=kafka_exporter,pid=43716,uid=65534\n",
    ]


@pytest.fixture
def sys_stdin_kubernetes_structured_burstable_systemd_cgroup():
    return [
        "some random line1\n",
        "1500316299 dev37-devc [30533610.306528] apache2 invoked oom-killer: "
        "gfp_mask=0x24000c0, order=0, oom_score_adj=0\n",
        "some random line2\n",
        "1500316300 dev37-devc [541471.893603] oom-kill:constraint=CONSTRAINT_MEMCG,"
        "nodemask=(null),cpuset=docker-e7ba37bd37089f8b1fda33c6f1fe753421ca6216518594bc73bca2ead7c13ba0.scope,"
        "mems_allowed=0,oom_memcg=/kubepods.slice/kubepods-burstable.slice/"
        "kubepods-burstable-pod8a0a7a03_d305_4ebc_83ad_91180c9d5ef9.slice/"
        "docker-e7ba37bd37089f8b1fda33c6f1fe753421ca6216518594bc73bca2ead7c13ba0.scope,"
        "task_memcg=/kubepods.slice/kubepods-burstable.slice/kubepods-burstable-pod8a0a7a03_d305_4ebc_83ad_91180c9d5ef9.slice/"
        "docker-e7ba37bd37089f8b1fda33c6f1fe753421ca6216518594bc73bca2ead7c13ba0.scope,task=apache2,pid=1757658,uid=0\n",
    ]


@pytest.fixture
def sys_stdin_kubernetes_containerd_systemd_cgroup_structured():
    return [
        "some random line1\n",
        "1720128512 dev37-devc [ 7195.442797] python3 invoked oom-killer: "
        "gfp_mask=0xcc0(GFP_KERNEL), order=0, oom_score_adj=999\n",
        "some random line2\n",
        "1720128512 dev37-devc [ 7195.442928] oom-kill:constraint=CONSTRAINT_MEMCG,"
        "cpuset=cri-containerd-e216d2f1e6c625d363c71edb6b3cbab5a9e1b447641b61028d0b94b077adf27c.scope,"
        "mems_allowed=0,oom_memcg=/kubepods.slice/kubepods-burstable.slice/"
        "kubepods-burstable-pod08768c36_163c_40e5_8e49_09cf42ff5046.slice,"
        "task_memcg=/kubepods.slice/kubepods-burstable.slice/kubepods-burstable-pod08768c36_163c_40e5_8e49_09cf42ff5046.slice/"
        "cri-containerd-e216d2f1e6c625d363c71edb6b3cbab5a9e1b447641b61028d0b94b077adf27c.scope,task=python3,pid=485850,uid=33\n",
    ]


@pytest.fixture
def sys_stdin_kubernetes_containerd_systemd_cgroup():
    return [
        "some random line1\n",
        "1720128512 dev208-uswest1adevc [42201.484624] python3 invoked oom-killer: "
        "gfp_mask=0xcc0(GFP_KERNEL), order=0, oom_score_adj=999\n",
        "some random line2\n",
        "1720128512 dev208-uswest1adevc [42201.484749] oom-kill:constraint=CONSTRAINT_MEMCG,"
        "nodemask=(null),cpuset=kubepods-burstable-pod73331cbb_9b96_4a62_9702_46a56ad49dd0.slice:"
        "cri-containerd:52f9ece9bcf929a08951aa3b4312fbec50890d82b58988f91a0aa9dc96ebc199,"
        "mems_allowed=0,oom_memcg=/system.slice/kubepods-burstable-pod73331cbb_9b96_4a62_9702_46a56ad49dd0.slice:"
        "cri-containerd:52f9ece9bcf929a08951aa3b4312fbec50890d82b58988f91a0aa9dc96ebc199,"
        "task_memcg=/system.slice/kubepods-burstable-pod73331cbb_9b96_4a62_9702_46a56ad49dd0.slice:"
        "cri-containerd:52f9ece9bcf929a08951aa3b4312fbec50890d82b58988f91a0aa9dc96ebc199,task=python3,pid=4190418,uid=33\n",
    ]


@pytest.fixture
def sys_stdin_process_name_with_slashes():
    return [
        "some random line1\n",
        "1500316299 dev37-devc [30533610.306528] /nail/live/yelp invoked oom-killer: "
        "gfp_mask=0x24000c0, order=0, oom_score_adj=0\n,"
        "some random line2\n",
        "1500316300 dev37-devc [30533610.306529] Task in "
        "/docker/a687af92e281725daf5b4cda0b487f20d2055d2bb6814b76d0e39c18a52a4e79 "
        "killed as a result of limit of "
        "/docker/a687af92e281725daf5b4cda0b487f20d2055d2bb6814b76d0e39c18a52a4e79\n",
    ]


@pytest.fixture
def sys_stdin_process_name_with_spaces():
    return [
        "some random line1\n",
        "1500316299 dev37-devc [30533610.306528] python batch/ke invoked oom-killer: "
        "gfp_mask=0x24000c0, order=0, oom_score_adj=0\n,"
        "some random line2\n",
        "1500316300 dev37-devc [30533610.306529] Task in "
        "/docker/a687af92e281725daf5b4cda0b487f20d2055d2bb6814b76d0e39c18a52a4e79 "
        "killed as a result of limit of "
        "/docker/a687af92e281725daf5b4cda0b487f20d2055d2bb6814b76d0e39c18a52a4e79\n",
    ]


@pytest.fixture
def sys_stdin_without_process_name():
    return [
        "some random line1\n",
        "1500216300 dev37-devc [1140036.678311] Task in "
        "/docker/e3a1057fdd485f5dffe48f1584e6f30c2bf6d30107d95518aea32bbb8bb29560 "
        "killed as a result of limit of "
        "/docker/e3a1057fdd485f5dffe48f1584e6f30c2bf6d30107d95518aea32bbb8bb29560\n,"
        "some random line2\n",
        "1500316300 dev37-devc [30533610.306529] Task in "
        "/docker/a687af92e281725daf5b4cda0b487f20d2055d2bb6814b76d0e39c18a52a4e79 "
        "killed as a result of limit of "
        "/docker/a687af92e281725daf5b4cda0b487f20d2055d2bb6814b76d0e39c18a52a4e79\n",
    ]


@pytest.fixture
def docker_inspect():
    return {
        "Config": {
            "Env": [
                "PAASTA_SERVICE=fake_service",
                "PAASTA_INSTANCE=fake_instance",
                "PAASTA_RESOURCE_MEM=512",
                "MESOS_CONTAINER_NAME=mesos-a04c14a6-83ea-4047-a802-92b850b1624e",
            ]
        }
    }


@pytest.fixture
def containerd_inspect():
    return {
        "process": {
            "env": [
                "PAASTA_SERVICE=fake_service",
                "PAASTA_INSTANCE=fake_instance",
                "PAASTA_RESOURCE_MEM=512",
            ]
        }
    }


@pytest.fixture
def log_line():
    return LogLine(
        timestamp=1500316300,
        hostname="dev37-devc",
        container_id="a687af92e281",
        cluster="fake_cluster",
        service="fake_service",
        instance="fake_instance",
        process_name="apache2",
        mesos_container_id="mesos-a04c14a6-83ea-4047-a802-92b850b1624e",
        mem_limit="512",
    )


@pytest.fixture
def log_line_containerd():
    return LogLine(
        timestamp=1720128512,
        hostname="dev37-devc",
        container_id="e216d2f1e6c625d363c71edb6b3cbab5a9e1b447641b61028d0b94b077adf27c",
        cluster="fake_cluster",
        service="fake_service",
        instance="fake_instance",
        process_name="python3",
        mesos_container_id="mesos-null",
        mem_limit="512",
    )


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
def test_capture_oom_events_from_stdin(mock_sys_stdin, sys_stdin):
    mock_sys_stdin.readline.side_effect = sys_stdin
    test_output = []
    for a_tuple in capture_oom_events_from_stdin():
        test_output.append(a_tuple)

    assert test_output == [(1500316300, "dev37-devc", "a687af92e281", "apache2")]


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
def test_capture_oom_events_from_stdin_kubernetes_qos(
    mock_sys_stdin,
    sys_stdin_kubernetes_besteffort_qos,
    sys_stdin_kubernetes_burstable_qos,
    sys_stdin_kubernetes_guaranteed_qos,
):
    for qos in (
        sys_stdin_kubernetes_besteffort_qos,
        sys_stdin_kubernetes_burstable_qos,
        sys_stdin_kubernetes_guaranteed_qos,
    ):
        mock_sys_stdin.readline.side_effect = qos
        test_output = []
        for a_tuple in capture_oom_events_from_stdin():
            test_output.append(a_tuple)
        assert test_output == [(1500316300, "dev37-devc", "0e4a814eda03", "apache2")]


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
def test_capture_oom_events_from_stdin_kubernetes_structured_qos(
    mock_sys_stdin,
    sys_stdin_kubernetes_structured_burstable_qos,
):
    mock_sys_stdin.readline.side_effect = sys_stdin_kubernetes_structured_burstable_qos
    test_output = []
    for a_tuple in capture_oom_events_from_stdin():
        test_output.append(a_tuple)
    assert test_output == [(1500316300, "dev37-devc", "0e4a814eda03", "apache2")]


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
def test_capture_oom_events_from_stdin_kubernetes_structured_burstable_systemd_cgroup(
    mock_sys_stdin,
    sys_stdin_kubernetes_structured_burstable_systemd_cgroup,
):
    mock_sys_stdin.readline.side_effect = (
        sys_stdin_kubernetes_structured_burstable_systemd_cgroup
    )
    test_output = [a_tuple for a_tuple in capture_oom_events_from_stdin()]
    assert test_output == [(1500316300, "dev37-devc", "e7ba37bd3708", "apache2")]


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
def test_capture_oom_events_from_stdin_kubernetes_containerd_systemd_cgroup(
    mock_sys_stdin,
    sys_stdin_kubernetes_containerd_systemd_cgroup,
):
    mock_sys_stdin.readline.side_effect = sys_stdin_kubernetes_containerd_systemd_cgroup
    test_output = [a_tuple for a_tuple in capture_oom_events_from_stdin()]
    assert test_output == [
        (
            1720128512,
            "dev208-uswest1adevc",
            "52f9ece9bcf929a08951aa3b4312fbec50890d82b58988f91a0aa9dc96ebc199",
            "python3",
        )
    ]


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
def test_capture_oom_events_from_stdin_kubernetes_containerd_systemd_cgroup_structured(
    mock_sys_stdin,
    sys_stdin_kubernetes_containerd_systemd_cgroup_structured,
):
    mock_sys_stdin.readline.side_effect = (
        sys_stdin_kubernetes_containerd_systemd_cgroup_structured
    )
    test_output = [a_tuple for a_tuple in capture_oom_events_from_stdin()]
    assert test_output == [
        (
            1720128512,
            "dev37-devc",
            "e216d2f1e6c625d363c71edb6b3cbab5a9e1b447641b61028d0b94b077adf27c",
            "python3",
        )
    ]


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
def test_capture_oom_events_from_stdin_with_slashes(
    mock_sys_stdin, sys_stdin_process_name_with_slashes
):
    mock_sys_stdin.readline.side_effect = sys_stdin_process_name_with_slashes
    test_output = []
    for a_tuple in capture_oom_events_from_stdin():
        test_output.append(a_tuple)

    assert test_output == [
        (1500316300, "dev37-devc", "a687af92e281", "/nail/live/yelp")
    ]


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
def test_capture_oom_events_from_stdin_with_spaces(
    mock_sys_stdin, sys_stdin_process_name_with_spaces
):
    mock_sys_stdin.readline.side_effect = sys_stdin_process_name_with_spaces
    test_output = []
    for a_tuple in capture_oom_events_from_stdin():
        test_output.append(a_tuple)

    assert test_output == [
        (1500316300, "dev37-devc", "a687af92e281", "python batch/ke")
    ]


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
def test_capture_oom_events_from_stdin_without_process_name(
    mock_sys_stdin, sys_stdin_without_process_name
):
    mock_sys_stdin.readline.side_effect = sys_stdin_without_process_name
    test_output = []
    for a_tuple in capture_oom_events_from_stdin():
        test_output.append(a_tuple)

    assert test_output == [
        (1500216300, "dev37-devc", "e3a1057fdd48", ""),
        (1500316300, "dev37-devc", "a687af92e281", ""),
    ]


@patch("paasta_tools.oom_logger.clog", autospec=True)
def test_log_to_clog(mock_clog, log_line):
    log_to_clog(log_line)
    mock_clog.log_line.assert_called_once_with(
        "tmp_paasta_oom_events",
        json.dumps(
            {
                "timestamp": log_line.timestamp,
                "hostname": log_line.hostname,
                "container_id": log_line.container_id,
                "cluster": log_line.cluster,
                "service": log_line.service,
                "instance": log_line.instance,
                "process_name": log_line.process_name,
                "mesos_container_id": log_line.mesos_container_id,
                "mem_limit": log_line.mem_limit,
            }
        ),
    )


@patch("paasta_tools.oom_logger.get_instance_config", autospec=True)
def test_send_sfx_event(mock_get_instance_config):
    service = "foo"
    instance = "bar"
    cluster = "baz"

    # Try to use the autospec if it's available
    from paasta_tools.oom_logger import yelp_meteorite

    if yelp_meteorite is None:
        autospec = None
    else:
        autospec = True

    with patch(
        "paasta_tools.oom_logger.yelp_meteorite", autospec=autospec
    ) as mock_meteorite:
        send_sfx_event(service, instance, cluster)

        expected_dimensions = {
            "paasta_service": service,
            "paasta_instance": instance,
            "paasta_cluster": cluster,
            "paasta_pool": mock_get_instance_config.return_value.get_pool.return_value,
        }
        mock_meteorite.events.emit_event.assert_called_once_with(
            "paasta.service.oom_events", dimensions=expected_dimensions
        )
        mock_meteorite.create_counter.assert_called_once_with(
            "paasta.service.oom_count", default_dimensions=expected_dimensions
        )
        assert mock_meteorite.create_counter.return_value.count.call_count == 1


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
@patch("paasta_tools.oom_logger.clog", autospec=True)
@patch("paasta_tools.oom_logger.send_sfx_event", autospec=True)
@patch("paasta_tools.oom_logger.load_system_paasta_config", autospec=True)
@patch("paasta_tools.oom_logger.log_to_clog", autospec=True)
@patch("paasta_tools.oom_logger.log_to_paasta", autospec=True)
@patch("paasta_tools.oom_logger.get_docker_client", autospec=True)
@patch("paasta_tools.oom_logger.parse_args", autospec=True)
def test_main(
    mock_parse_args,
    mock_get_docker_client,
    mock_log_to_paasta,
    mock_log_to_clog,
    mock_load_system_paasta_config,
    mock_send_sfx_event,
    mock_clog,
    mock_sys_stdin,
    sys_stdin,
    docker_inspect,
    log_line,
):

    mock_sys_stdin.readline.side_effect = sys_stdin
    mock_parse_args.return_value.containerd = False
    docker_client = Mock(inspect_container=Mock(return_value=docker_inspect))
    mock_get_docker_client.return_value = docker_client
    mock_load_system_paasta_config.return_value.get_cluster.return_value = (
        "fake_cluster"
    )

    main()
    mock_log_to_paasta.assert_called_once_with(log_line)
    mock_log_to_clog.assert_called_once_with(log_line)
    mock_send_sfx_event.assert_called_once_with(
        "fake_service", "fake_instance", "fake_cluster"
    )


@patch("paasta_tools.oom_logger.sys.stdin", autospec=True)
@patch("paasta_tools.oom_logger.clog", autospec=True)
@patch("paasta_tools.oom_logger.send_sfx_event", autospec=True)
@patch("paasta_tools.oom_logger.load_system_paasta_config", autospec=True)
@patch("paasta_tools.oom_logger.log_to_clog", autospec=True)
@patch("paasta_tools.oom_logger.log_to_paasta", autospec=True)
@patch("paasta_tools.oom_logger.parse_args", autospec=True)
@patch("paasta_tools.oom_logger.get_containerd_container", autospec=True)
@patch("paasta_tools.oom_logger.json.loads", autospec=True)
def test_main_containerd(
    mock_json_loads,
    mock_get_containerd_container,
    mock_parse_args,
    mock_log_to_paasta,
    mock_log_to_clog,
    mock_load_system_paasta_config,
    mock_send_sfx_event,
    mock_clog,
    mock_sys_stdin,
    sys_stdin_kubernetes_containerd_systemd_cgroup_structured,
    log_line_containerd,
    containerd_inspect,
):

    mock_sys_stdin.readline.side_effect = (
        sys_stdin_kubernetes_containerd_systemd_cgroup_structured
    )
    mock_parse_args.return_value.containerd = True

    mock_container_info = MagicMock()
    mock_container_info.spec.value.decode.return_value = str(containerd_inspect)

    mock_get_containerd_container.return_value = mock_container_info
    mock_json_loads.return_value = containerd_inspect

    mock_load_system_paasta_config.return_value.get_cluster.return_value = (
        "fake_cluster"
    )

    main()
    mock_log_to_paasta.assert_called_once_with(log_line_containerd)
    mock_log_to_clog.assert_called_once_with(log_line_containerd)
    mock_send_sfx_event.assert_called_once_with(
        "fake_service", "fake_instance", "fake_cluster"
    )



import mock
import pytest

from paasta_tools.list_kubernetes_service_instances import main
from paasta_tools.list_kubernetes_service_instances import parse_args


def test_parse_args():
    with mock.patch(
        "paasta_tools.list_kubernetes_service_instances.argparse.ArgumentParser",
        autospec=True,
    ) as mock_parser:
        assert parse_args() == mock_parser.return_value.parse_args()


@mock.patch("paasta_tools.list_kubernetes_service_instances.parse_args", autospec=True)
@mock.patch(
    "paasta_tools.list_kubernetes_service_instances.get_services_for_cluster",
    autospec=True,
    return_value=[("service_1", "instance1"), ("service_2", "instance1")],
)
@mock.patch("builtins.print", autospec=True)
@pytest.mark.parametrize(
    "instance_type,sanitise,expected,shuffle,group_lines",
    [
        ("kubernetes", False, "service_1.instance1\nservice_2.instance1", False, None),
        ("kubernetes", True, "service--1-instance1\nservice--2-instance1", False, None),
        ("flink", True, "service--1-instance1\nservice--2-instance1", False, None),
        ("eks", False, "service_1.instance1\nservice_2.instance1", False, None),
        ("eks", True, "service--1-instance1\nservice--2-instance1", False, None),
    ],
)
def test_main(
    mock_print,
    mock_get_services,
    mock_parse_args,
    instance_type,
    sanitise,
    expected,
    shuffle,
    group_lines,
):
    mock_parse_args.return_value = mock.Mock(
        instance_type=instance_type,
        sanitise=sanitise,
        shuffle=shuffle,
        group_lines=group_lines,
    )

    with pytest.raises(SystemExit) as e:
        main()

    assert e.value.code == 0
    assert mock_get_services.call_args_list == [
        mock.call(
            cluster=mock_parse_args.return_value.cluster,
            instance_type=instance_type,
            soa_dir=mock_parse_args.return_value.soa_dir,
        )
    ]
    assert mock_print.call_args_list == [mock.call(expected)]



import os
from subprocess import CalledProcessError

import mock
import pytest

from paasta_tools.apply_external_resources import main


@pytest.fixture
def mock_run():
    with mock.patch(
        "paasta_tools.apply_external_resources.run", autospec=True
    ) as mock_runner:
        yield mock_runner


@pytest.fixture(autouse=True)
def setup_external_files(fs):
    fs.create_file(
        "/external_resources/00-common/10-foo/10-deployment.yaml",
        contents="foo: bar",
    )
    fs.create_file(
        "/external_resources/00-common/10-foo/20-service.yaml",
        contents="fizz: buzz",
    )
    fs.create_file(
        "/external_resources/20-common/10-foo/20-deployment.yaml",
        contents="baz: biz",
    )
    fs.create_file(
        "/external_resources/.applied/00-common/10-foo/10-deployment.yaml",
        contents="foo: bar",
    )
    fs.create_file(
        "/external_resources/.applied/00-common/10-foo/20-service.yaml",
        contents="fizz: buzz",
    )
    fs.create_file(
        "/external_resources/.applied/20-common/10-foo/20-deployment.yaml",
        contents="baz: biz",
    )


def test_no_changes(mock_run):
    assert main("/external_resources") == 0
    assert mock_run.call_count == 0


def test_resources_added_in_order(mock_run, fs):
    fs.create_file(
        "/external_resources/00-common/10-foo/30-hpa.yaml",
        contents="blah: blah",
    )
    fs.create_file(
        "/external_resources/00-common/10-foo/40-service.yaml",
        contents="blah: blah",
    )
    fs.create_file(
        "/external_resources/00-common/30-foo/10-deployment.yaml",
        contents="blah: blah",
    )
    assert main("/external_resources") == 0
    assert mock_run.call_args_list == [
        mock.call(
            [
                "kubectl",
                "apply",
                "-f",
                "/external_resources/00-common/10-foo/30-hpa.yaml",
            ],
            check=True,
        ),
        mock.call(
            [
                "kubectl",
                "apply",
                "-f",
                "/external_resources/00-common/10-foo/40-service.yaml",
            ],
            check=True,
        ),
        mock.call(
            [
                "kubectl",
                "apply",
                "-f",
                "/external_resources/00-common/30-foo/10-deployment.yaml",
            ],
            check=True,
        ),
    ]
    assert os.path.exists("/external_resources/.applied/00-common/10-foo/30-hpa.yaml")
    assert os.path.exists(
        "/external_resources/.applied/00-common/10-foo/40-service.yaml"
    )
    assert os.path.exists(
        "/external_resources/.applied/00-common/30-foo/10-deployment.yaml"
    )


def test_resources_deleted_in_reverse_order(mock_run, fs):
    fs.create_file(
        "/external_resources/.applied/00-common/10-foo/30-hpa.yaml",
        contents="blah: blah",
    )
    fs.create_file(
        "/external_resources/.applied/00-common/10-foo/40-service.yaml",
        contents="blah: blah",
    )
    assert main("/external_resources") == 0
    assert mock_run.call_args_list == [
        mock.call(
            [
                "kubectl",
                "delete",
                "--ignore-not-found=true",
                "-f",
                "/external_resources/.applied/00-common/10-foo/40-service.yaml",
            ],
            check=True,
        ),
        mock.call(
            [
                "kubectl",
                "delete",
                "--ignore-not-found=true",
                "-f",
                "/external_resources/.applied/00-common/10-foo/30-hpa.yaml",
            ],
            check=True,
        ),
    ]
    assert not os.path.exists(
        "/external_resources/.applied/00-common/10-foo/30-hpa.yaml"
    )
    assert not os.path.exists(
        "/external_resources/.applied/00-common/10-foo/40-service.yaml"
    )


def test_kubectl_fails(mock_run, fs):
    mock_run.side_effect = [CalledProcessError(cmd="kubectl", returncode=1), None]
    fs.create_file(
        "/external_resources/00-common/10-foo/30-hpa.yaml",
        contents="blah: blah",
    )
    fs.create_file(
        "/external_resources/00-common/10-foo/40-service.yaml",
        contents="blah: blah",
    )
    assert main("/external_resources") == 1
    assert not os.path.exists(
        "/external_resources/.applied/00-common/10-foo/30-hpa.yaml"
    )
    assert os.path.exists(
        "/external_resources/.applied/00-common/10-foo/40-service.yaml"
    )



import socket
from contextlib import contextmanager

import mock
import pytest

from paasta_tools import docker_wrapper
from paasta_tools import docker_wrapper_imports


@contextmanager
def patch_environ(envs):
    with mock.patch(
        "paasta_tools.docker_wrapper.os.environ", new=envs, autospec=None
    ) as environ:
        yield environ


class ImmutableDict(dict):
    def __setitem__(self, key, value):
        pass


class TestParseEnvArgs:
    @pytest.fixture(autouse=True)
    def mock_empty_os_environ(self):
        # Patch `os.environ` to contain a not changeable empty set of environment variables.
        with patch_environ(ImmutableDict()) as environ:
            yield environ

    def test_empty(self):
        env = docker_wrapper.parse_env_args(["docker"])
        assert env == {}

    @pytest.mark.parametrize(
        "args",
        [
            ["docker", "-e", "key=value"],
            ["docker", "-e=key=value"],
            ["docker", "-te", "key=value"],  # -e can be combined with other short args
            [
                "docker",
                "-et",
                "key=value",
            ],  # -t takes no additional parameters so docker allows this order
        ],
    )
    def test_short(self, args):
        env = docker_wrapper.parse_env_args(args)
        assert env == {"key": "value"}

    @pytest.mark.parametrize(
        "args,envs",
        [
            (["docker", "-e", "key"], {"key": "value"}),
            (["docker", "-e=key"], {"key": "value"}),
            (["docker", "-te", "key"], {"key": "value"}),
            (["docker", "-et", "key"], {"key": "value"}),
        ],
    )
    def test_short_with_envs(self, args, envs):
        with patch_environ(envs):
            env = docker_wrapper.parse_env_args(args)
            assert env == {"key": "value"}

    @pytest.mark.parametrize(
        "args", [["docker", "--env", "key=value"], ["docker", "--env=key=value"]]
    )
    def test_long(self, args):
        env = docker_wrapper.parse_env_args(args)
        assert env == {"key": "value"}

    @pytest.mark.parametrize(
        "args,envs",
        [
            (["docker", "--env", "key"], {"key": "value"}),
            (["docker", "--env=key=value"], {"key": "value"}),
        ],
    )
    def test_long_with_envs(self, args, envs):
        with patch_environ(envs):
            env = docker_wrapper.parse_env_args(args)
            assert env == {"key": "value"}

    @pytest.mark.parametrize(
        "args",
        [
            [
                "docker",
                "--eep",
                "key=value",
            ],  # ignore -e in double dashes unless its --env
            ["docker", "-a", "key=value"],  # only -e should matter
            ["docker", "-"],  # just don't crash
        ],
    )
    def test_short_invalid(self, args):
        env = docker_wrapper.parse_env_args(args)
        assert env == {}

    def test_mixed_short_long(self):
        env = docker_wrapper.parse_env_args(
            ["docker", "-e", "foo=bar", "--env=apple=banana", "--env", "c=d"]
        )
        assert env == {"foo": "bar", "apple": "banana", "c": "d"}

    def test_multiple_equals(self):
        env = docker_wrapper.parse_env_args(["docker", "-e", "foo=bar=cat"])
        assert env == {"foo": "bar=cat"}

    def test_dupe(self):
        env = docker_wrapper.parse_env_args(
            ["docker", "-e", "foo=bar", "-e", "foo=cat"]
        )
        assert env == {"foo": "cat"}

    def test_empty_value(self):
        env = docker_wrapper.parse_env_args(["docker", "-e", "foo=", "--env=bar="])
        assert env == {"foo": "", "bar": ""}

    def test_file_equals(self, mock_env_file):
        env = docker_wrapper.parse_env_args(["docker", f"--env-file={mock_env_file}"])
        assert env == {"fileKeyA": "fileValueA", "fileKeyB": "fileValueB"}

    def test_file(self, mock_env_file):
        env = docker_wrapper.parse_env_args(["docker", "--env-file", mock_env_file])
        assert env == {"fileKeyA": "fileValueA", "fileKeyB": "fileValueB"}

    def test_two_files(self, mock_env_file, mock_env_file2):
        env = docker_wrapper.parse_env_args(
            ["docker", "--env-file", mock_env_file, "--env-file", mock_env_file2]
        )
        assert env == {
            "fileKeyA": "fileValueA",
            "fileKeyB": "fileValueB",
            "fileKeyC": "fileValueC",
        }

    def test_file_and_short(self, mock_env_file):
        env = docker_wrapper.parse_env_args(
            ["docker", "--env-file", mock_env_file, "-e", "foo=bar"]
        )
        assert env == {"foo": "bar", "fileKeyA": "fileValueA", "fileKeyB": "fileValueB"}

    @pytest.fixture
    def mock_env_file(self, tmpdir):
        env_file = tmpdir.join("env.txt")
        env_file.write("fileKeyA=fileValueA\nfileKeyB=fileValueB\n")
        return str(env_file)

    @pytest.fixture
    def mock_env_file2(self, tmpdir):
        env_file = tmpdir.join("env2.txt")
        env_file.write("fileKeyC=fileValueC\n")
        return str(env_file)


class TestCanAddHostname:
    def test_empty(self):
        assert docker_wrapper.can_add_hostname(["docker"]) is True

    @pytest.mark.parametrize(
        "args", [["docker", "-h"], ["docker", "-th"], ["docker", "-ht"]]
    )
    def test_short(self, args):
        assert docker_wrapper.can_add_hostname(args) is False

    @pytest.mark.parametrize(
        "args",
        [
            ["docker", "-e=foo=hhh"],
            ["docker", "--hhh"],
            ["docker", "-"],
            ["docker", "--net", "bridged"],
            ["docker", "--net"],
            ["docker", "--network", "bridged"],
            ["docker", "--network"],
        ],
    )
    def test_short_invalid(self, args):
        assert docker_wrapper.can_add_hostname(args) is True

    @pytest.mark.parametrize(
        "args",
        [
            ["docker", "--hostname", "foo"],
            ["docker", "--hostname=foo"],
            ["docker", "--net=host"],
            ["docker", "--net", "host"],
            ["docker", "--network=host"],
            ["docker", "--network", "host"],
            ["docker", "--hostname=foo", "--net=host"],
        ],
    )
    def test_long(self, args):
        assert docker_wrapper.can_add_hostname(args) is False


class TestCanAddMacAddress:
    def test_empty(self):
        assert docker_wrapper.can_add_mac_address(["docker"]) is False

    def test_run(self):
        assert docker_wrapper.can_add_mac_address(["docker", "run"]) is True

    def test_not_run(self):
        assert docker_wrapper.can_add_mac_address(["docker", "inspect"]) is False

    @pytest.mark.parametrize(
        "args",
        [
            ["docker", "--mac-address", "foo"],
            ["docker", "--mac-address=foo"],
            ["docker", "--net=host"],
            ["docker", "--net", "host"],
            ["docker", "--network=host"],
            ["docker", "--network", "host"],
            ["docker", "--mac-address=foo", "--net=host"],
        ],
    )
    def test_long(self, args):
        assert docker_wrapper.can_add_mac_address(args) is False


class TestGenerateHostname:
    def test_simple(self):
        hostname = docker_wrapper.generate_hostname_task_id(
            "first", "what.only.matters.is.lastpart"
        )
        assert hostname == "first-lastpart"

    def test_truncate(self):
        hostname = docker_wrapper.generate_hostname_task_id(
            "reallllllllllllllylooooooooooooooong",
            "reallyreallylongidsssssssssssssssssssssssss",
        )
        assert (
            hostname == "reallllllllllllllylooooooooooooooong-reallyreallylongidsssss"
        )
        assert len(hostname) == 60

    def test_symbols(self):
        hostname = docker_wrapper.generate_hostname_task_id(
            "first", "anything:can_do!s0me weird-stuff"
        )
        assert hostname == "first-anything-can-do-s0me-weird-stuff"

    def test_no_dashes_on_end(self):
        assert (
            docker_wrapper.generate_hostname_task_id("beep", "foobar-") == "beep-foobar"
        )

        hostname = docker_wrapper.generate_hostname_task_id(
            "reallllllllllllllylooooooooooooooong", "reallyreallylongid0123--abc"
        )
        assert hostname == "reallllllllllllllylooooooooooooooong-reallyreallylongid0123"
        assert len(hostname) == 59


@pytest.mark.parametrize(
    "input_args,expected_args",
    [
        (["docker", "ps"], ["docker", "ps"]),  # do not add it for non-run commands
        (  # add it after the first run arg
            ["docker", "run", "run"],
            ["docker", "run", "--hostname=myhostname", "run"],
        ),
        (  # ignore args before run
            ["docker", "--host", "/fake.sock", "run", "-t"],
            ["docker", "--host", "/fake.sock", "run", "--hostname=myhostname", "-t"],
        ),
    ],
)
def test_add_argument(input_args, expected_args):
    args = docker_wrapper.add_argument(input_args, "--hostname=myhostname")
    assert args == expected_args


class TestMain:
    @pytest.fixture(autouse=True)
    def mock_execlp(self):
        # always patch execlp so we don't actually exec
        with mock.patch.object(
            docker_wrapper.os, "execlp", autospec=True
        ) as mock_execlp:
            yield mock_execlp

    def test_marathon(self, mock_execlp):
        argv = [
            "docker",
            "run",
            "--env=MESOS_TASK_ID=paasta--canary.main.git332d4a22.config458863b1.0126a188-f944-11e6-bdfb-12abac3adf8c",
        ]
        with mock.patch.object(socket, "getfqdn", return_value="myhostname"):
            docker_wrapper.main(argv)
        assert mock_execlp.mock_calls == [
            mock.call(
                "docker",
                "docker",
                "run",
                "--hostname=myhostname-0126a188-f944-11e6-bdfb-12abac3adf8c",
                "-e=PAASTA_HOST=myhostname",
                "--env=MESOS_TASK_ID=paasta--canary.main.git332d4a22.config458863b1.0126a188-f944-11e6-bdfb-12abac3adf8c",
            )
        ]

    def test_env_not_present(self, mock_execlp):
        argv = ["docker", "run", "foobar"]
        docker_wrapper.main(argv)
        assert mock_execlp.mock_calls == [
            mock.call(
                "docker",
                "docker",
                "run",
                f"--hostname={socket.gethostname()}",
                f"-e=PAASTA_HOST={socket.getfqdn()}",
                "foobar",
            )
        ]

    def test_already_has_hostname(self, mock_execlp):
        argv = [
            "docker",
            "run",
            "--env=MESOS_TASK_ID=my-mesos-task-id",
            "--hostname=somehostname",
        ]
        with mock.patch.object(socket, "getfqdn", return_value="myhostname"):
            docker_wrapper.main(argv)
        assert mock_execlp.mock_calls == [
            mock.call(
                "docker",
                "docker",
                "run",
                "--env=MESOS_TASK_ID=my-mesos-task-id",
                "--hostname=somehostname",
            )
        ]

    def test_not_run(self, mock_execlp):
        argv = ["docker", "ps", "--env=MESOS_TASK_ID=my-mesos-task-id"]
        docker_wrapper.main(argv)
        assert mock_execlp.mock_calls == [
            mock.call("docker", "docker", "ps", "--env=MESOS_TASK_ID=my-mesos-task-id")
        ]

    @pytest.fixture
    def mock_mac_address(self):
        with mock.patch.object(
            docker_wrapper_imports,
            "reserve_unique_mac_address",
            return_value=("00:00:00:00:00:00", None),
        ) as mock_mac_address:
            yield mock_mac_address

    @pytest.fixture
    def mock_firewall_env_args(self):
        return [
            "--env=PAASTA_FIREWALL=1",
            "--env=PAASTA_SERVICE=myservice",
            "--env=PAASTA_INSTANCE=myinstance",
        ]

    @mock.patch.object(docker_wrapper_imports, "firewall_flock", autospec=True)
    @mock.patch.object(docker_wrapper_imports, "prepare_new_container", autospec=True)
    def test_mac_address(
        self,
        mock_prepare_new_container,
        mock_firewall_flock,
        mock_mac_address,
        mock_execlp,
        mock_firewall_env_args,
    ):
        argv = ["docker", "run"] + mock_firewall_env_args
        docker_wrapper.main(argv)

        assert mock_execlp.mock_calls == [
            mock.call(
                "docker",
                "docker",
                "run",
                "--mac-address=00:00:00:00:00:00",
                f"--hostname={socket.gethostname()}",
                f"-e=PAASTA_HOST={socket.getfqdn()}",
                *mock_firewall_env_args,
            )
        ]

        assert mock_firewall_flock.return_value.__enter__.called is True

        assert mock_prepare_new_container.mock_calls == [
            mock.call(
                docker_wrapper_imports.DEFAULT_SOA_DIR,
                docker_wrapper_imports.DEFAULT_SYNAPSE_SERVICE_DIR,
                "myservice",
                "myinstance",
                "00:00:00:00:00:00",
            )
        ]

    def test_mac_address_not_run(
        self, mock_mac_address, mock_execlp, mock_firewall_env_args
    ):
        argv = ["docker", "inspect"] + mock_firewall_env_args
        docker_wrapper.main(argv)

        assert mock_mac_address.call_count == 0
        assert mock_execlp.mock_calls == [
            mock.call("docker", "docker", "inspect", *mock_firewall_env_args)
        ]

    def test_mac_address_already_set(
        self, mock_mac_address, mock_execlp, mock_firewall_env_args
    ):
        argv = [
            "docker",
            "run",
            "--mac-address=12:34:56:78:90:ab",
        ] + mock_firewall_env_args
        docker_wrapper.main(argv)

        assert mock_mac_address.call_count == 0
        assert mock_execlp.mock_calls == [
            mock.call(
                "docker",
                "docker",
                "run",
                f"--hostname={socket.gethostname()}",
                f"-e=PAASTA_HOST={socket.getfqdn()}",
                "--mac-address=12:34:56:78:90:ab",
                *mock_firewall_env_args,
            )
        ]

    def test_mac_address_no_lockdir(
        self, capsys, mock_execlp, tmpdir, mock_firewall_env_args
    ):
        nonexistent = tmpdir.join("nonexistent")
        with mock.patch.object(docker_wrapper, "LOCK_DIRECTORY", str(nonexistent)):
            argv = ["docker", "run"] + mock_firewall_env_args
            docker_wrapper.main(argv)

            assert mock_execlp.mock_calls == [
                mock.call(
                    "docker",
                    "docker",
                    "run",
                    f"--hostname={socket.gethostname()}",
                    f"-e=PAASTA_HOST={socket.getfqdn()}",
                    *mock_firewall_env_args,
                )
            ]
            _, err = capsys.readouterr()
            assert err.startswith(
                "Unable to add mac address: [Errno 2] No such file or directory"
            )

    @mock.patch.object(
        docker_wrapper_imports,
        "firewall_flock",
        autospec=True,
        side_effect=Exception("Oh noes"),
    )
    @mock.patch.object(docker_wrapper_imports, "prepare_new_container", autospec=True)
    def test_prepare_new_container_error(
        self,
        mock_prepare_new_container,
        mock_firewall_flock,
        capsys,
        mock_mac_address,
        mock_execlp,
        mock_firewall_env_args,
    ):
        argv = ["docker", "run"] + mock_firewall_env_args
        docker_wrapper.main(argv)

        assert mock_execlp.mock_calls == [
            mock.call(
                "docker",
                "docker",
                "run",
                "--mac-address=00:00:00:00:00:00",
                f"--hostname={socket.gethostname()}",
                f"-e=PAASTA_HOST={socket.getfqdn()}",
                *mock_firewall_env_args,
            )
        ]
        _, err = capsys.readouterr()
        assert err.startswith("Unable to add firewall rules: Oh noes")



# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# We just want to test that task_processing is available in the virtualenv


def test_import():
    from task_processing.task_processor import TaskProcessor

    tp = TaskProcessor()
    tp.load_plugin("task_processing.plugins.mesos")
